/*-
 * Copyright (c) 2013 Chris Rees <crees@FreeBSD.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <cfuconf.h>
#include <err.h>
#include <errno.h>
#include <libgen.h>
#include <libutil.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "keyboard-daemon.h"

static cfuconf_t *conf;
static struct pidfh *pfh;
static FILE *dev;

/*
 * Use libcfu to read the config file, map scan codes to commands
 * and execute them
 */
int
main(int argc, char *argv[])
{
	pid_t fpid;
	int ch, dopt, vopt;
	char *buf, *configfile, *device, *key, *pidfile;

	configfile = pidfile = NULL;
	pfh = NULL;
	dopt = vopt = 0;

	while ((ch = getopt(argc, argv, "Vdf:p:v")) != -1) {
		switch (ch) {
		case 'V':
			printf("keyboard-daemon-%s\n", VERSION);
			exit(0);
			break;
		case 'd':
			dopt = 1;
			break;
		case 'f':
			configfile = realpath(optarg, NULL);
			break;
		case 'p':
			pidfile = optarg;
			break;
		case 'v':
			vopt = 1;
			break;
		default:
			usage();
			break;
		}
	}

	argc -= optind;
	argv += optind;

	/* Only one remaining argument permitted; taken as device name */
	if (argc != 1)
		usage();

	device = argv[0];


	if (configfile == NULL)
		configfile = realpath(DEFAULT_CONFIG_FILE, NULL);

	/*
	 * There is a race here of course, but we are only warning because
	 * we care about the user.
	 */
	if (check_permissions(configfile) != 0)
		err(EXIT_FAILURE, "While checking permissions on %s", configfile);

	if (cfuconf_parse_file(configfile, &conf, &buf) != 0) {
		if (buf != NULL) {
			warnx("While reading config file \"%s\": %s\n", 
			    configfile, buf);
			free(buf);
			exit(1);
		} else {
			errx(1, "Something unknown failed while opening the config file");
		}
	}

	if (dopt == 1) {
		if (pidfile != NULL) {
			pfh = pidfile_open(pidfile, 0600, &fpid);
			if (pfh == NULL) {
				if (errno == EEXIST)
					errx(EXIT_FAILURE,
					    "keyboard-daemon already running, pid: %d.",
					    fpid);
				err(EXIT_FAILURE, "Cannot open pidfile");
			}
		}

		if (daemon(0, 0) == -1) {
			warn("Cannot daemonize");
			pidfile_remove(pfh);
			exit(EXIT_FAILURE);
		}

		pidfile_write(pfh);
	}

	if ((dev = fopen(device, "r")) == NULL)
		err(1, "%s", device);

	while ((ch = getc(dev)) != EOF) {
		key = int_to_hexstring(ch);
		if (cfuconf_get_directive_one_arg(conf, key, &buf) == 0) {
			/*
			 * The input here is trusted; only root should
			 * be able to write to this config file
			 */
			fpid = fork();
			if (fpid == 0) {
				system(buf);
				exit(0);
			} else if (fpid == -1)
				err(1, "forking failed");
		} else if (vopt) {
			warnx("Unrecognised scan code: %s", key);
		}
		free(key);
	}
	
	exit(0);
}

/* Check permissions all the way back to root "this function is recursive" */
int check_permissions(char *conffile) {
	struct stat mode;

	/* We have reached root */
	if (strlen(conffile) == 1) {
		if (conffile[0] == '/')
			return (0);
		else {
			/*
			 * This should never be reached as
			 * realpath is called on conffile
			 */
			errno = EINVAL;
			warnx("-f must be an absolute path");
			return (1);
		}
	}

	if (stat(conffile, &mode) != 0)
		return (1);

	if (mode.st_nlink > 1 && S_ISDIR(mode.st_mode) == 0) {
		warnx("Multiple links to %s not allowed", conffile);
		errno = EMLINK;
		return (1);
	}

	if (mode.st_uid != 0) {
		warnx("%s is not owned by root, DANGEROUS AND INSECURE!",
		      conffile);
		errno = EACCES;
		return (1);
	}

	/* Owner must be root, group may not write */
	if ((mode.st_mode & (S_IWGRP | S_IWOTH)) != 0) {
		warnx("%s is writable by !root, DANGEROUS AND INSECURE!",
		      conffile);
		errno = EACCES;
		return (1);
	}

	return (check_permissions(dirname(conffile)));
}

/*
 * Returns a string containing a hexadecimal representation of a passed int.
 * Extra cautious here on string length; must be able to contain the maximum
 * possible length of an int, so
 * length = log16(2 ^ sizeof(int) * 8) - 1 = sizeof(int) * 2 - 1
 *
 * Free it!
 */
char *
int_to_hexstring(int ch)
{
	size_t l;
	char *text;

	l = sizeof(int) * 2;
	if ((text = malloc(l)) == NULL)
		return (NULL);

	snprintf(text, l - 1, "%x", ch);

	return (text);
}

static void
usage(void)
{

	fprintf(stderr, "usage: [-Vdv] [-f conffile] [-p pidfile] device\n");
	exit(1);
}
